@@ -89,7 +89,7 @@ static void set_name_value(editorconfig_name_value* nv, const char* name,
             !strcmp(nv->name, "charset"))
         strlwr(nv->value);
 
-    /* set speical pointers */
+    /* set special pointers */
     set_special_property_name_value_pointers(nv, spnvp);
 }
 
@@ -138,7 +138,7 @@ static int array_editorconfig_name_value_add(
     int         name_value_pos;
     /* always use name_lwr but not name, since property names are case
      * insensitive */
-    char        name_lwr[MAX_PROPERTY_NAME];
+    char        name_lwr[MAX_PROPERTY_NAME+1] = {0};
     /* For the first time we came here, aenv->name_values is NULL */
     if (aenv->name_values == NULL) {
         aenv->name_values = (editorconfig_name_value*)malloc(
@@ -153,7 +153,7 @@ static int array_editorconfig_name_value_add(
 
 
     /* name_lwr is the lowercase property name */
-    strlwr(strcpy(name_lwr, name));
+    strlwr(strncpy(name_lwr, name, MAX_PROPERTY_NAME));
 
     name_value_pos = find_name_value_from_name(
             aenv->name_values, aenv->current_value_count, name_lwr);
@@ -228,16 +228,34 @@ static int ini_handler(void* hfp, const char* section, const char* name,
         return 1;
     }
 
-    /* pattern would be: /dir/of/editorconfig/file[double_star]/[section] if
+    /* Pattern would be: /dir/of/editorconfig/file[double_star]/[section] if
      * section does not contain '/', or /dir/of/editorconfig/file[section]
      * if section starts with a '/', or /dir/of/editorconfig/file/[section] if
-     * section contains '/' but does not start with '/' */
+     * section contains '/' but does not start with '/'.
+     *
+     * If the dir part has any special characters as defined by ec_glob.c, we
+     * need to escape them.
+     */
     pattern = (char*)malloc(
-            strlen(hfparam->editorconfig_file_dir) * sizeof(char) +
+        /* The 2 here is for possible escaping. */
+        strlen(hfparam->editorconfig_file_dir) * sizeof(char) * 2 +
             sizeof("**/") + strlen(section) * sizeof(char));
     if (!pattern)
         return 0;
-    strcpy(pattern, hfparam->editorconfig_file_dir);
+
+    /* Escaping special characters in the directory part. */
+    char* ptr = hfparam->editorconfig_file_dir;
+    char* ptr_prev = ptr;
+    char* ptr_pattern = pattern;
+    for (; (ptr = strpbrk(ptr, ec_special_chars)) != NULL; ++ ptr, ptr_prev = ptr)
+    {
+        ptrdiff_t s = ptr - ptr_prev;
+        memcpy(ptr_pattern, ptr_prev, s * sizeof(char));
+        ptr_pattern += s;
+        *(ptr_pattern ++) = '\\';  /* escaping char */
+        *(ptr_pattern ++) = *ptr;
+    }
+    strcpy(ptr_pattern, ptr_prev);
 
     if (strchr(section, '/') == NULL) /* No / is found, append '[star][star]/' */
         strcat(pattern, "**/");
@@ -259,13 +277,15 @@ static int ini_handler(void* hfp, const char* section, const char* name,
     return 1;
 }
 
-/* 
+/*
  * Split an absolute file path into directory and filename parts.
  *
  * If absolute_path does not contain a path separator, set directory and
  * filename to NULL pointers.
- */ 
-static void split_file_path(char** directory, char** filename,
+ *
+ * Returns -1 if an OOM error occurs. Otherwise returns 0.
+ */
+static int split_file_path(char** directory, char** filename,
         const char* absolute_path)
 {
     char* path_char = strrchr(absolute_path, '/');
@@ -275,16 +295,26 @@ static void split_file_path(char** directory, char** filename,
             *directory = NULL;
         if (filename)
             *filename = NULL;
-        return;
+        return 0;
     }
 
     if (directory != NULL) {
         *directory = strndup(absolute_path,
                 (size_t)(path_char - absolute_path));
+        if (*directory == NULL)
+            return -1;
     }
+
     if (filename != NULL) {
         *filename = strndup(path_char+1, strlen(path_char)-1);
+        if (*filename == NULL) {
+            if (directory != NULL)
+                free(*directory);
+            return -1;
+        }
     }
+
+    return 0;
 }
 
 /*
@@ -303,23 +333,31 @@ static int count_slashes(const char* filename)
 
 /*
  * Return an array of full filenames for files in every directory in and above
- * the given path with the name of the relative filename given.
+ * the given path with the name of the relative filename given. Return NULL if
+ * failed (OOM).
  */
 static char** get_filenames(const char* path, const char* filename)
 {
-    char* currdir;
-    char* currdir1;
+    char* currdir = NULL;
     char** files;
     int slashes = count_slashes(path);
     int i;
 
     files = (char**) calloc(slashes+1, sizeof(char*));
+    if (files == NULL)
+        goto failure_cleanup;
 
     currdir = strdup(path);
+    if (currdir == NULL)
+        goto failure_cleanup;
+
     for (i = slashes - 1; i >= 0; --i) {
-        currdir1 = currdir;
-        split_file_path(&currdir, NULL, currdir1);
+        char* currdir1 = currdir;
+        int err_split;
+        err_split = split_file_path(&currdir, NULL, currdir1);
         free(currdir1);
+        if (err_split == -1)
+            goto failure_cleanup;
         files[i] = malloc(strlen(currdir) + strlen(filename) + 2);
         strcpy(files[i], currdir);
         strcat(files[i], "/");
@@ -331,6 +369,32 @@ static char** get_filenames(const char* path, const char* filename)
     files[slashes] = NULL;
 
     return files;
+
+failure_cleanup:
+
+    free(currdir);
+
+    if (files != NULL) {
+        for (i = 0; i < slashes; ++ i)
+            free(files[i]);
+        free(files);
+    }
+
+    return NULL;
+}
+
+/*
+ * Free the memory used by an array of strings that was created by
+ * get_filenames().
+ */
+static void free_filenames(char **filenames)
+{
+    if (filenames != NULL) {
+        for (char** filename = filenames; *filename != NULL; filename++) {
+            free(*filename);
+        }
+        free(filenames);
+    }
 }
 
 /*
@@ -381,7 +445,7 @@ const char* editorconfig_get_error_msg(int err_num)
     return "Unknown error.";
 }
 
-/* 
+/*
  * See the header file for the use of this function
  */
 EDITORCONFIG_EXPORT
@@ -389,8 +453,8 @@ int editorconfig_parse(const char* full_filename, editorconfig_handle h)
 {
     handler_first_param                 hfp;
     char**                              config_file;
-    char**                              config_files;
-    int                                 err_num;
+    char**                              config_files = NULL;
+    int                                 err_num = 0;
     int                                 i;
     struct editorconfig_handle*         eh = (struct editorconfig_handle*)h;
     struct editorconfig_version         cur_ver;
@@ -409,7 +473,7 @@ int editorconfig_parse(const char* full_filename, editorconfig_handle h)
     if (editorconfig_compare_version(&eh->ver, &cur_ver) > 0)
         return EDITORCONFIG_PARSE_VERSION_TOO_NEW;
 
-    if (!eh->err_file) {
+    if (eh->err_file) {
         free(eh->err_file);
         eh->err_file = NULL;
     }
@@ -433,10 +497,15 @@ int editorconfig_parse(const char* full_filename, editorconfig_handle h)
     memset(&hfp, 0, sizeof(hfp));
 
     hfp.full_filename = strdup(full_filename);
+    if (hfp.full_filename == NULL) {
+        err_num = EDITORCONFIG_PARSE_MEMORY_ERROR;
+        goto cleanup;
+    }
 
     /* return an error if file path is not absolute */
     if (!is_file_path_absolute(full_filename)) {
-        return EDITORCONFIG_PARSE_NOT_FULL_PATH;
+        err_num = EDITORCONFIG_PARSE_NOT_FULL_PATH;
+        goto cleanup;
     }
 
 #ifdef WIN32
@@ -447,20 +516,31 @@ int editorconfig_parse(const char* full_filename, editorconfig_handle h)
     array_editorconfig_name_value_init(&hfp.array_name_value);
 
     config_files = get_filenames(hfp.full_filename, eh->conf_file_name);
+    if (config_files == NULL) {
+        err_num = EDITORCONFIG_PARSE_MEMORY_ERROR;
+        goto cleanup;
+    }
     for (config_file = config_files; *config_file != NULL; config_file++) {
-        split_file_path(&hfp.editorconfig_file_dir, NULL, *config_file);
-        if ((err_num = ini_parse(*config_file, ini_handler, &hfp)) != 0 &&
+        int ini_err_num;
+        err_num = split_file_path(&hfp.editorconfig_file_dir, NULL, *config_file);
+        if (err_num == -1) {
+          err_num = EDITORCONFIG_PARSE_MEMORY_ERROR;
+          goto cleanup;
+        }
+
+        if ((ini_err_num = ini_parse(*config_file, ini_handler, &hfp)) != 0 &&
                 /* ignore error caused by I/O, maybe caused by non exist file */
-                err_num != -1) {
+                ini_err_num != -1) {
+            /* No need to specifically deal with the return value of the strdup
+               of this line. If any error occurs for this strdup call,
+               eh->err_file would simply be NULL.*/
             eh->err_file = strdup(*config_file);
-            free(*config_file);
-            free(hfp.full_filename);
-            free(hfp.editorconfig_file_dir);
-            return err_num;
+            err_num = ini_err_num;
+            goto cleanup;
         }
 
         free(hfp.editorconfig_file_dir);
-        free(*config_file);
+        hfp.editorconfig_file_dir = NULL;
     }
 
     /* value proprocessing */
@@ -498,7 +578,7 @@ int editorconfig_parse(const char* full_filename, editorconfig_handle h)
 
     if (eh->name_value_count == 0) {  /* no value is set, just return 0. */
         free(hfp.full_filename);
-        free(config_files);
+        free_filenames(config_files);
         return 0;
     }
     eh->name_values = hfp.array_name_value.name_values;
@@ -507,13 +587,16 @@ int editorconfig_parse(const char* full_filename, editorconfig_handle h)
             sizeof(editorconfig_name_value) * eh->name_value_count);
     if (eh->name_values == NULL) {
         free(hfp.full_filename);
+        free_filenames(config_files);
         return EDITORCONFIG_PARSE_MEMORY_ERROR;
     }
 
+ cleanup:
+    free_filenames(config_files);
     free(hfp.full_filename);
-    free(config_files);
+    free(hfp.editorconfig_file_dir);
 
-    return 0;
+    return err_num;
 }
 
 /*
@@ -523,11 +606,11 @@ EDITORCONFIG_EXPORT
 void editorconfig_get_version(int* major, int* minor, int* patch)
 {
     if (major)
-        *major = editorconfig_VERSION_MAJOR;
+        *major = EC_VERSION_MAJOR;
     if (minor)
-        *minor = editorconfig_VERSION_MINOR;
+        *minor = EC_VERSION_MINOR;
     if (patch)
-        *patch = editorconfig_VERSION_PATCH;
+        *patch = EC_VERSION_PATCH;
 }
 
 /*
@@ -536,5 +619,5 @@ void editorconfig_get_version(int* major, int* minor, int* patch)
 EDITORCONFIG_EXPORT
 const char* editorconfig_get_version_suffix(void)
 {
-    return editorconfig_VERSION_SUFFIX;
+    return EC_VERSION_SUFFIX;
 }
